/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver;

import com.aquenos.epics.jackie.client.ClientThreadingStrategy;
import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.resolver.internal.DefaultChannelNameResolverImpl;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;

/**
 * <p>
 * Default implementation of a {@link ChannelNameResolver}.
 * </p>
 * 
 * <p>
 * This class provides an implementation that should be suitable for most use
 * cases, as most important parameters can be configured by through a
 * {@link ChannelNameResolverConfiguration}. It can search for channels through
 * both TCP and UDP and implements a listener for server beacons.
 * </p>
 * 
 * <p>
 * In order to work, an object of this class needs to be registered with a
 * {@link CommunicationController}. Typically, code using this class will not
 * instantiate the communication controller directly but instead use a
 * {@link ClientThreadingStrategy} which will create the appropriate controller
 * and pass it to this class when it is registered through the
 * {@link ClientThreadingStrategy#setChannelNameResolver(CommunicationProcessor)
 * setChannelNameResolver(...)} method.
 * </p>
 * 
 * <p>
 * This implementation tries to stay as close to the implementation in the C
 * library as reasonably possible - as far as the implementation details have an
 * effect on the network behavior (e.g. timing of packets).
 * </p>
 * 
 * <p>
 * As far as the user-side is concerned, this class is thread safe. This means
 * that the methods provided by the {@link ChannelNameResolver} interface can be
 * called from any thread, at any time, even before this object has been
 * registered with a communication controller. However, calling those methods
 * will have no effect until the registration with the communication controller
 * is completed.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class DefaultChannelNameResolver implements ChannelNameResolver,
        CommunicationProcessor {

    private DefaultChannelNameResolverImpl implementation;

    /**
     * Creates a channel name resolver that uses the specified configuration.
     * The resolver has to be associated with a {@link CommunicationController}
     * before it is fully functional.
     * 
     * @param configuration
     *            channel-name resolver configuration (never <code>null</code>).
     * @param beaconDetector
     *            beacon detector used by this channel name resolver or
     *            <code>null</code> if no beacon detector should be used. The
     *            channel name resolver uses the beacon detector for two
     *            purposes: First, when a beacon anomaly has occurred, it tries
     *            to resolve channel names that it could not resolve in the past
     *            at a higher interval. Second, when an irregular beacon is
     *            received for a server that it uses for name resolution via
     *            TCP, it checks whether the connection is still working.
     */
    public DefaultChannelNameResolver(
            ChannelNameResolverConfiguration configuration,
            BeaconDetector beaconDetector) {
        this.implementation = new DefaultChannelNameResolverImpl(configuration,
                beaconDetector);
    }

    @Override
    public void resolveChannelName(String channelName, int cid,
            ChannelNameSearchResultProcessor resultProcessor) {
        implementation.resolveChannelName(channelName, cid, resultProcessor);
    }

    @Override
    public void resolveChannelName(String channelName, int cid,
            ChannelNameSearchResultProcessor resultProcessor, long initialDelay) {
        implementation.resolveChannelName(channelName, cid, resultProcessor,
                initialDelay);
    }

    @Override
    public void cancelChannelName(int cid) {
        implementation.cancelChannelName(cid);

    }

    @Override
    public void setCommunicationController(
            CommunicationController communicationController) {
        implementation.setCommunicationController(communicationController);
    }

    /**
     * Destroys this channel name resolver. This closes the I/O channels opened
     * by this resolver. The {@link CommunicationController} that is running the
     * communication thread should be stopped before calling this method.
     * Otherwise, unexpected exceptions might be thrown in the communication
     * thread. The channel name resolver is not usable after calling this method
     * and should be discarded.
     */
    public void destroy() {
        implementation.destroy();
    }

}
