/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver.internal;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.nio.channels.SocketChannel;
import java.util.HashMap;

import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolverConfiguration;
import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;
import com.aquenos.epics.jackie.common.io.TimerProcessor;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSearchClientMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSearchTCPServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;

/**
 * <p>
 * Channel name resolver for the TCP based protocol.
 * </p>
 * 
 * <p>
 * This resolver uses non-blocking {@link SocketChannel}s for sending and
 * receiving TCP packets. One channel is created for each configured name
 * server. Received messages are processed immediately. Sent messages might be
 * queued until they can be sent.
 * </p>
 * 
 * <p>
 * This class is only intended for use by {@link DefaultChannelNameResolverImpl}
 * . It is closely tied to that class and has only been separated into a class
 * of its own to keep that class smaller.
 * </p>
 * 
 * <p>
 * This class is not thread safe, but it is only called by its
 * {@link CommunicationController} or methods of other classes that are managed
 * by the same {@link CommunicationController}.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class TcpResolver implements CommunicationProcessor {

    private BeaconDetector beaconDetector;
    private CommunicationController communicationController;
    private ChannelNameResolverConfiguration configuration;
    private HashMap<InetSocketAddress, TcpNameServerConnection> connections = new HashMap<InetSocketAddress, TcpNameServerConnection>();
    private TimerProcessor connectionTimer = new TimerProcessor() {
        @Override
        public void processTimer() {
            connect();
        }
    };
    private volatile boolean destroyed = false;
    private HashMap<InetSocketAddress, Long> lastConnectionAttemptTime = new HashMap<InetSocketAddress, Long>();
    private HashMap<InetSocketAddress, Long> nextConnectionAttemptTime = new HashMap<InetSocketAddress, Long>();
    private DefaultChannelNameResolverImpl resolver;

    /**
     * Creates a TCP resolver that is owned by <code>resolver</code>.
     * 
     * @param configuration
     *            channel-name resolver configuration.
     * @param resolver
     *            channel name resolver owning this resolver.
     * @param beaconDetector
     *            beacon detector used to assess the health of TCP connections
     *            or <code>null</code> if beacons should not be considered.
     */
    public TcpResolver(ChannelNameResolverConfiguration configuration,
            DefaultChannelNameResolverImpl resolver,
            BeaconDetector beaconDetector) {
        this.configuration = configuration;
        this.resolver = resolver;
        this.beaconDetector = beaconDetector;
    }

    /**
     * Tries to connect to all servers.
     */
    private void connect() {
        for (InetSocketAddress serverAddress : configuration.getTcpAddresses()) {
            connect(serverAddress);
        }
    }

    /**
     * Tries to connect to a specific server. If a connection attempt is pending
     * or the connection has already been established, nothing is done. If a
     * connection attempt for the server has failed recently, no connection
     * attempt is made.
     * 
     * @param serverAddress
     *            address of the server for which a connection should be
     *            established.
     */
    private void connect(InetSocketAddress serverAddress) {
        // If the TCP resolver has been destroyed, we do not want to create new
        // connections.
        if (destroyed) {
            return;
        }
        // We check for an existing connection before trying to create a new
        // one.
        TcpNameServerConnection existingConnection = connections
                .get(serverAddress);
        if (existingConnection != null && !existingConnection.isDestroyed()) {
            // The connection is connected or a connection attempt is pending.
            // We do not have to create a new connection.
            return;
        }
        // The connection has been disconnected or has never been connected. We
        // check whether we should try to connect it.
        Long nextConnectionAttemptTime = this.nextConnectionAttemptTime
                .get(serverAddress);
        if (nextConnectionAttemptTime != null
                && nextConnectionAttemptTime > System.currentTimeMillis()) {
            // The time for the next attempt has not come yet, thus we discard
            // the connection for the moment.
            return;
        } else {
            this.nextConnectionAttemptTime.remove(serverAddress);
        }
        // Either the connection does not exist yet or it has been destroyed and
        // is now ready to be reconnected.
        this.lastConnectionAttemptTime.put(serverAddress,
                System.currentTimeMillis());
        TcpNameServerConnection connection;
        try {
            connection = new TcpNameServerConnection(this, beaconDetector,
                    configuration.getCharset(), communicationController,
                    configuration.getErrorHandler(), serverAddress);
        } catch (SocketException e) {
            // If the connection fails right away, we want to wait a few seconds
            // before we try again.
            long nextTime = System.currentTimeMillis() + 15000L;
            this.nextConnectionAttemptTime.put(serverAddress, nextTime);
            communicationController.registerTimer(connectionTimer, nextTime);
            return;
        } catch (IOException e) {
            // If the connection fails right away, we want to wait a few seconds
            // before we try again.
            long nextTime = System.currentTimeMillis() + 15000L;
            this.nextConnectionAttemptTime.put(serverAddress, nextTime);
            communicationController.registerTimer(connectionTimer, nextTime);
            // We want to log any IOException that is not a SocketException.
            configuration
                    .getErrorHandler()
                    .handleError(
                            TcpResolver.class,
                            e,
                            "Unexpected I/O exception while trying to create a connection to a TCP name server.");
            return;
        }
        connections.put(serverAddress, connection);
    }

    /**
     * Called by a {@link TcpNameServerConnection} when a response to a search
     * request is received.
     * 
     * @param message
     *            search response received.
     * @param serverVersion
     *            Channel Access version supported by the server sending the
     *            response.
     */
    void receivedSearchResponse(ChannelAccessSearchTCPServerMessage message,
            ChannelAccessVersion serverVersion) {
        resolver.processSearchResponse(message, serverVersion);
    }

    /**
     * Called by a {@link TcpNameServerConnection} when it becomes available.
     * 
     * @param serverAddress
     *            IP address and port number of the server for which the
     *            connection is now available.
     */
    void serverAvailable(InetSocketAddress serverAddress) {
        // If no server has been available before, we want to wake up the
        // resolver so that it can now process its timers. If a different server
        // has been available before, this will cause one unnecessary timer
        // processing, but this does not hurt.
        resolver.scheduleImmediateTimerProcessing();
    }

    /**
     * Called by a {@link TcpNameServerConnection} when it is destroyed. This
     * method may be called by any thread, not just the communication thread.
     * However, calling it has no effect when the calling thread is not the
     * communication thread.
     * 
     * @param serverAddress
     *            IP address and port number of the server for which the
     *            connection has been destroyed.
     */
    void serverDestroyed(InetSocketAddress serverAddress) {
        // If the connection was destroyed explicitly we might not be in the
        // communication controller thread. In this case, we do not want to do
        // anything.
        if (!communicationController.inCommunicationThread()) {
            return;
        }
        Long lastConnectionAttemptTime = this.lastConnectionAttemptTime
                .get(serverAddress);
        // If the connection existed for more than five seconds, we try to
        // reconnect immediately. Otherwise, we block the next connection
        // attempt for fifteen seconds in order to avoid a high rate of
        // connect attempts.
        if (lastConnectionAttemptTime == null
                || lastConnectionAttemptTime + 5000L < System
                        .currentTimeMillis()) {
            connect(serverAddress);
        } else {
            long nextConnectionAttemptTime = System.currentTimeMillis() + 15000L;
            this.nextConnectionAttemptTime.put(serverAddress,
                    nextConnectionAttemptTime);
            communicationController.registerTimer(connectionTimer,
                    nextConnectionAttemptTime);
        }
    }

    @Override
    public void setCommunicationController(
            CommunicationController communicationController) {
        this.communicationController = communicationController;
        if (communicationController.inCommunicationThread()) {
            connect();
        } else {
            communicationController.registerTimer(new TimerProcessor() {
                @Override
                public void processTimer() {
                    connect();
                }
            }, 0L);
        }
    }

    /**
     * Tells whether at least one server connection is available and can process
     * search requests.
     * 
     * @return <code>true</code> if at least one server connection is available,
     *         <code>false</code> if no connection is available (there are
     *         either no connections or all connections are not available).
     */
    public boolean isAnyServerAvailable() {
        for (TcpNameServerConnection connection : connections.values()) {
            if (connection.isAvailable()) {
                return true;
            }
        }
        return false;
    }

    /**
     * Sends a search message.
     * 
     * @param cid
     *            numeric, client-assigned channel ID. This ID is used to
     *            identify the channel in a response to the search request.
     * @param channelName
     *            channel name.
     * @throws SerializedChannelNameTooLargeException
     *             if the channel name is too long to be serialized.
     */
    public void sendSearchMessage(int cid, String channelName) {
        ChannelAccessSearchClientMessage searchMessage = new ChannelAccessSearchClientMessage(
                cid, ChannelAccessVersion.NEWEST_SUPPORTED_VERSION, false,
                channelName);
        // We check the server connections and try to connect those connections
        // that are not connected yet.
        connect();
        // Send the search message over all connections that are available.
        for (TcpNameServerConnection connection : connections.values()) {
            if (connection.isAvailable()) {
                connection.sendSearchMessage(searchMessage);
            }
        }
    }

    /**
     * Destroys this TCP resolver. This also destroys all TCP connections. This
     * method should be called when the resolver is no longer needed. This
     * method may be called by any thread, not just the communication thread.
     */
    public void destroy() {
        destroyed = true;
        for (TcpNameServerConnection connection : connections.values()) {
            // The destroy method may be called from any thread, not just the
            // communication thread.
            connection.destroy();
        }
    }

}
