/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.ShortPayloadException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Representation of a <code>CA_PROTO_EVENT_ADD</code> message that is sent from
 * a server to a client. This message is sent in response to a subscription
 * cancellation request.
 * </p>
 * 
 * <p>
 * For some historic reasons, the server sends this message instead of a
 * <code>CA_PROTO_EVENT_CANCEL</code> message in response to a
 * {@link ChannelAccessCancelSubscriptionClientMessage}. The only way to tell
 * this message apart from a {@link ChannelAccessSubscriptionServerMessage} is
 * by its payload size. This message always has a zero payload size, while the
 * {@link ChannelAccessSubscriptionServerMessage} always has a non-zero payload
 * size.
 * </p>
 * 
 * @see ChannelAccessCommand#CA_PROTO_EVENT_ADD
 * @see ChannelAccessSubscriptionClientMessage
 * @see ChannelAccessSubscriptionServerMessage
 * @see ChannelAccessCancelSubscriptionClientMessage
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessCancelSubscriptionServerMessage extends
        ChannelAccessSubscriptionMessage {

    private short dataType;
    private int count;
    private int sid;
    private int subscriptionId;

    /**
     * Creates a <code>CA_PROTO_EVENT_ADD</code> message. This constructor
     * should be used to create a message that is sent from a server to a client
     * in response to a {@link ChannelAccessCancelSubscriptionClientMessage}
     * sent by the client.
     * 
     * @param dataType
     *            data type sent in the client's request.
     * @param count
     *            count sent in the client's request.
     * @param sid
     *            the channel identifier used by the server.
     * @param subscriptionId
     *            client-assigned subscription ID. This ID must be the same ID
     *            that was sent with the client's request.
     */
    public ChannelAccessCancelSubscriptionServerMessage(
            ChannelAccessValueType dataType, int count, int sid,
            int subscriptionId) {
        super();

        this.dataType = dataType.toTypeCode();
        this.count = count;
        this.sid = sid;
        this.subscriptionId = subscriptionId;
    }

    private ChannelAccessCancelSubscriptionServerMessage(
            ChannelAccessMessageHeader messageHeader, short dataType,
            int count, int sid, int subscriptionId) {
        super(messageHeader);
        this.dataType = dataType;
        this.count = count;
        this.sid = sid;
        this.subscriptionId = subscriptionId;
    }

    /**
     * Returns the data type of the value in this message's payload. This should
     * always match the data type that the client requested. Returns
     * <code>null</code> if the specified data-type is unknown to this
     * implementation.
     * 
     * @return data-type of the value or <code>null</code> if the numeric
     *         identifier cannot be mapped to a data type known by this
     *         implementation.
     */
    public ChannelAccessValueType getDataType() {
        try {
            return ChannelAccessValueType.forTypeNumber(dataType);
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    /**
     * Returns the number of elements the value in this message's payload has.
     * This should always match the element count that the client requested.
     * 
     * @return number of value elements.
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the channel ID that is used by the server to identify the
     * channel. This integer may be zero or negative.
     * 
     * @return server-side ID for the channel.
     */
    public int getChannelSID() {
        return sid;
    }

    /**
     * Returns the client-assigned subscription ID. The subscription ID is used
     * by the client in order to map events sent in response to a subscription
     * to the respective subscription. This must always be the ID that was sent
     * by the client. This number might be negative or zero.
     * 
     * @return client-assigned subscription identifier.
     */
    public int getSubscriptionId() {
        return subscriptionId;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, dataType, count, sid, subscriptionId,
                version, maxPayloadSize);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        verifyHeader(0, count, version, maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessCancelSubscriptionServerMessage other = (ChannelAccessCancelSubscriptionServerMessage) obj;
        return new EqualsBuilder().append(this.dataType, other.dataType)
                .append(this.count, other.count).append(this.sid, other.sid)
                .append(this.subscriptionId, other.subscriptionId).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(dataType).append(count).append(sid)
                .append(subscriptionId).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("dataType=");
        try {
            sb.append(ChannelAccessValueType.forTypeNumber(dataType));
        } catch (IllegalArgumentException e) {
            sb.append("UNKNOWN(");
            sb.append(dataType);
            sb.append(')');
        }
        sb.append(", count=");
        sb.append(count);
        sb.append(", sid=");
        sb.append(sid);
        sb.append(", subscriptionId=");
        sb.append(subscriptionId);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     * @throws ShortPayloadException
     *             if the message's payload size is less than the size needed
     *             for the data type and count.
     */
    protected static ChannelAccessCancelSubscriptionServerMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_EVENT_ADD
                .getCommandNumber());
        // The payload size must be zero. If it is not, a
        // ChannelAccessSubscriptionServerMessage should be created instead.
        assert (messageHeader.getPayloadSize() == 0);
        return new ChannelAccessCancelSubscriptionServerMessage(messageHeader,
                messageHeader.getDataType(), messageHeader.getCount(),
                messageHeader.getCID(), messageHeader.getContextSpecific());
    }
}
