/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.InvalidPriorityException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_VERSION</code> message that is sent from a
 * client to a server via TCP. This message is usually send at the start of a
 * new TCP connection in order to communicate the protocol version supported by
 * the client and set the priority that should be associated with this
 * connection. However, some servers might not consider the specified priority.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_VERSION
 * @see ChannelAccessVersionUDPMessage
 * @see ChannelAccessVersionTCPServerMessage
 */
public final class ChannelAccessVersionTCPClientMessage extends
        ChannelAccessVersionMessage {

    private short priority;

    /**
     * Creates a <code>CA_PROTO_VERSION</code> message. This constructor should
     * be used to create a message that is supposed to be sent from a client to
     * a server using the TCP protocol.
     * 
     * @param version
     *            Channel Access version supported by the client.
     * @param priority
     *            priority that should be associated with this TCP connection.
     *            This number must be between
     *            {@link ChannelAccessConstants#MIN_PRIORITY} and
     *            {@link ChannelAccessConstants#MAX_PRIORITY}. Higher numbers
     *            mean a higher priority.
     * @throws InvalidPriorityException
     *             if the specified priority is out of range.
     */
    public ChannelAccessVersionTCPClientMessage(ChannelAccessVersion version,
            short priority) {
        super(version);

        if (version == null) {
            throw new NullPointerException(
                    "Channel Acccess version must not be null.");
        }

        if (priority < ChannelAccessConstants.MIN_PRIORITY
                || priority > ChannelAccessConstants.MAX_PRIORITY) {
            throw new InvalidPriorityException();
        }
        this.priority = priority;
    }

    private ChannelAccessVersionTCPClientMessage(
            ChannelAccessMessageHeader messageHeader,
            ChannelAccessVersion version, short priority) {
        super(messageHeader, version);
        if (priority < ChannelAccessConstants.MIN_PRIORITY
                || priority > ChannelAccessConstants.MAX_PRIORITY) {
            throw new InvalidPriorityException();
        }
        this.priority = priority;
    }

    /**
     * Returns the priority requested by the client.
     * 
     * @return the priority requested by the client. This is a value between
     *         {@link ChannelAccessConstants#MIN_PRIORITY} and
     *         {@link ChannelAccessConstants#MAX_PRIORITY}.
     */
    public short getPriority() {
        return priority;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, priority,
                this.version.getMinorVersionAsShort(), 0, 0, version,
                maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessVersionTCPClientMessage other = (ChannelAccessVersionTCPClientMessage) obj;
        return this.priority == other.priority;
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(priority).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("version=");
        sb.append(version);
        sb.append(", priority=");
        sb.append(priority);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     * @throws InvalidPriorityException
     *             if the priority specified in the message header is out of
     *             range.
     */
    protected static ChannelAccessVersionTCPClientMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_VERSION
                .getCommandNumber());
        if (!headerOnly) {
            byteSource.skip(messageHeader.getPayloadSize());
        }
        return new ChannelAccessVersionTCPClientMessage(messageHeader,
                ChannelAccessVersion.forMinorVersion(messageHeader.getCount()),
                messageHeader.getDataType());
    }

}
