/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;
import com.aquenos.epics.jackie.common.value.ChannelAccessValue;

/**
 * Base class for Channel Access types that have strings as their main values.
 * 
 * @param <ImplementationType>
 *            type of the actual implementation.
 * 
 * @author Sebastian Marsching
 */
public abstract class AbstractStringBase<ImplementationType extends AbstractStringBase<ImplementationType>>
        extends ChannelAccessValueBase<String, ImplementationType> {

    private class WrappedStringList extends AbstractList<String> {

        @Override
        public String get(int index) {
            return value.get(index);
        }

        @Override
        public int size() {
            return value.size();
        }

        @Override
        public String set(int index, String element) {
            if (element == null) {
                throw new NullPointerException();
            }
            String old = value.set(index, element);
            updateRawValue(index, element);
            return old;
        }

        @Override
        public void add(int index, String element) {
            if (element == null) {
                throw new NullPointerException();
            }
            value.add(index, element);
            int newSize = value.size();
            adjustRawValueSize(newSize);
            for (int stringIndex = index; stringIndex < newSize; stringIndex++) {
                updateRawValue(stringIndex, value.get(stringIndex));
            }
        }

        @Override
        public String remove(int index) {
            String removed = value.remove(index);
            int newSize = value.size();
            for (int stringIndex = index; stringIndex < newSize; stringIndex++) {
                updateRawValue(stringIndex, value.get(stringIndex));
            }
            return removed;
        }

        // Overridden for an optimized implementation.
        @Override
        public void clear() {
            value.clear();
        }

        // Overridden for an optimized implementation.
        @Override
        public boolean addAll(Collection<? extends String> c) {
            return addAll(value.size(), c);
        }

        // Overridden for an optimized implementation.
        @Override
        public boolean addAll(int index, Collection<? extends String> c) {
            for (String s : c) {
                if (s == null) {
                    throw new NullPointerException();
                }
            }
            boolean modified = value.addAll(index, c);
            if (modified) {
                int newSize = value.size();
                adjustRawValueSize(newSize);
                for (int stringIndex = index; stringIndex < newSize; stringIndex++) {
                    updateRawValue(stringIndex, value.get(stringIndex));
                }
            }
            return modified;
        }

        // Overridden for an optimized implementation.
        @Override
        public Object[] toArray() {
            return value.toArray();
        }

        // Overridden for an optimized implementation.
        @Override
        public <T> T[] toArray(T[] a) {
            return value.toArray(a);
        }

    };

    private static final int EPICS_STRING_SIZE = 40;
    private static final byte[] EMPTY_STRING_BYTES = new byte[EPICS_STRING_SIZE];

    protected ArrayList<String> value;

    /**
     * Raw representation of the strings stored in <code>value</code>.
     */
    protected byte[] rawValue;

    /**
     * Encoding for strings.
     */
    protected Charset charset;

    /**
     * Base constructor.
     * 
     * @param value
     *            string value.
     * @param charset
     *            encoding used for encoding/decoding strings.
     */
    AbstractStringBase(Collection<String> value, Charset charset) {
        assert (value != null);
        assert (charset != null);
        this.charset = charset;
        int numberOfElements = value.size();
        this.rawValue = new byte[numberOfElements * EPICS_STRING_SIZE];
        this.value = new ArrayList<String>(numberOfElements);
        int stringIndex = 0;
        for (String element : value) {
            if (element == null) {
                throw new NullPointerException();
            }
            this.value.add(element);
            updateRawValue(stringIndex, element);
            stringIndex++;
        }
    }

    /**
     * Base constructor for read-only variant.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    AbstractStringBase(ImplementationType delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     * 
     * @param charset
     *            encoding used for encoding/decoding strings.
     */
    AbstractStringBase(Charset charset) {
        assert (charset != null);
        this.charset = charset;
    }

    @Override
    public int getValueSize() {
        if (delegate != null) {
            return delegate.getValueSize();
        }
        return value.size();
    }

    public List<String> getValue() {
        if (delegate != null) {
            return Collections.unmodifiableList(delegate.getValue());
        }
        return new WrappedStringList();
    }

    public void setValue(Collection<String> value) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        // We check for null pointers first, so that we do not modify the
        // elements list if there are any.
        for (String element : value) {
            if (element == null) {
                throw new NullPointerException();
            }
        }
        this.value.clear();
        adjustRawValueSize(value.size());
        int stringIndex = 0;
        for (String element : value) {
            this.value.add(element);
            updateRawValue(stringIndex, element);
            stringIndex++;
        }
    }

    @Override
    public String getGenericValueElement(int index) {
        if (delegate != null) {
            return delegate.getGenericValueElement(index);
        }
        return value.get(index);
    }

    private void adjustRawValueSize(int numberOfElements) {
        int minimumSize = numberOfElements * EPICS_STRING_SIZE;
        if (rawValue.length < minimumSize) {
            rawValue = Arrays.copyOf(rawValue, minimumSize);
        }
    }

    private void updateRawValue(int stringIndex, String newValue) {
        // The caller is responsible for resizing the rawValue array if
        // necessary.
        assert ((stringIndex + 1) * EPICS_STRING_SIZE <= rawValue.length);
        byte[] rawString = NullTerminatedStringUtil
                .stringToTruncatedNullTerminatedFixedBytes(newValue,
                        EPICS_STRING_SIZE, charset);
        System.arraycopy(rawString, 0, this.rawValue, stringIndex
                * EPICS_STRING_SIZE, EPICS_STRING_SIZE);
    }

    public byte[] getRawValue() {
        if (delegate != null) {
            return delegate.getRawValue();
        }
        return Arrays.copyOf(rawValue, value.size() * EPICS_STRING_SIZE);
    }

    public void setRawValue(byte[] rawValue) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        // Check that all strings have exactly 40 bytes.
        if (rawValue.length % EPICS_STRING_SIZE != 0) {
            throw new IllegalArgumentException(
                    "The length of the string's raw-value must be an integer multiple of 40.");
        }
        int numberOfStrings = rawValue.length / EPICS_STRING_SIZE;
        // Check that all strings are terminated by a null byte. We have to
        // remember the position of the null byte because we will need it when
        // we create the Java string objects.
        LinkedList<String> newStrings = new LinkedList<String>();
        for (int stringIndex = 0; stringIndex < numberOfStrings; stringIndex++) {
            newStrings.add(NullTerminatedStringUtil.nullTerminatedBytesToString(
                    rawValue, stringIndex * EPICS_STRING_SIZE,
                    EPICS_STRING_SIZE, charset, true));
        }
        // If the number of strings is less than or equal to the old number of
        // strings, we can reuse the raw array, otherwise, we have to create a
        // new one. In every case, we copy the parameter array.
        if (this.rawValue.length >= rawValue.length) {
            System.arraycopy(rawValue, 0, this.rawValue, 0, rawValue.length);
        } else {
            this.rawValue = rawValue.clone();
        }
        // We remove all existing strings from the value list and add the new
        // ones.
        this.value.clear();
        this.value.addAll(newStrings);
    }

    public Charset getCharset() {
        if (delegate != null) {
            return delegate.getCharset();
        }
        return charset;
    }

    /**
     * Deserializes the value elements from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @param count
     *            number of elements to deserialize.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    protected void deserializeValue(ByteSource byteSource, int count) {
        assert (count >= 0);
        value = new ArrayList<String>(count);
        if (count == 0) {
            byteSource.getByteArray(EPICS_STRING_SIZE);
            rawValue = ArrayUtils.EMPTY_BYTE_ARRAY;
        } else {
            rawValue = byteSource.getByteArray(EPICS_STRING_SIZE * count);
            for (int stringIndex = 0; stringIndex < count; stringIndex++) {
                // Ensure that the string is null terminated.
                rawValue[(stringIndex + 1) * EPICS_STRING_SIZE - 1] = 0;
                value.add(NullTerminatedStringUtil.nullTerminatedBytesToString(
                        rawValue, stringIndex * EPICS_STRING_SIZE,
                        EPICS_STRING_SIZE, charset, true));
            }
        }
    }

    /**
     * Serializes the (raw) value elements to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write to.
     * @param count
     *            number of elements to be serialized.
     */
    protected void serializeValue(ByteSink byteSink, int count) {
        assert (count >= 0);
        assert (count <= value.size());
        if (count == 0) {
            byteSink.putByteArray(EMPTY_STRING_BYTES);
        } else {
            byteSink.putByteArray(rawValue, 0, EPICS_STRING_SIZE * count);
        }
    }

    @Override
    public boolean equals(Object obj) {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been
        // checked.
        @SuppressWarnings("unchecked")
        AbstractStringBase<ImplementationType> other = (AbstractStringBase<ImplementationType>) obj;
        // The raw values might have different sizes, because not all bytes
        // might be used. Therefore we have to do the comparison by ourselves.
        if (!this.value.equals(other.value)) {
            return false;
        }
        if (!this.charset.equals(other.charset)) {
            return false;
        }
        // We already compared the values, therefore we know that both raw
        // values must be of appropriate size.
        int rawSize = this.value.size() * EPICS_STRING_SIZE;
        for (int byteIndex = 0; byteIndex < rawSize; byteIndex++) {
            if (this.rawValue[byteIndex] != other.rawValue[byteIndex]) {
                return false;
            }
        }
        return true;
    }

    @Override
    public int hashCode() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        HashCodeBuilder hcb = new HashCodeBuilder()
                .appendSuper(super.hashCode()).append(value).append(charset);
        // The raw value might not be used completely, therefore we cannot
        // append the whole array.
        int rawSize = value.size() * EPICS_STRING_SIZE;
        for (int byteIndex = 0; byteIndex < rawSize; byteIndex++) {
            hcb.append(rawValue[byteIndex]);
        }
        return hcb.toHashCode();
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessValue<String> clone() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        AbstractStringBase<ImplementationType> copy = (AbstractStringBase<ImplementationType>) super
                .clone();
        copy.value = (ArrayList<String>) this.value.clone();
        copy.rawValue = this.rawValue.clone();
        return copy;
    }

}
