/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.ShortBuffer;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessConfigureAcknowledgeTransientAlarms;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Implementation of DBR_PUT_ACKT.
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessConfigureAcknowledgeTransientAlarmsImpl
        extends
        ChannelAccessValueBase<Short, ChannelAccessConfigureAcknowledgeTransientAlarmsImpl>
        implements ChannelAccessConfigureAcknowledgeTransientAlarms {

    protected short[] value;

    /**
     * Creates a DBR_PUT_ACKT.
     * 
     * @param explicitlyAcknowledgeTransientAlarms
     *            Do transient alarms have to be acknowledged explicitly?
     */
    public ChannelAccessConfigureAcknowledgeTransientAlarmsImpl(
            boolean explicitlyAcknowledgeTransientAlarms) {
        this.value = new short[] { (short) (explicitlyAcknowledgeTransientAlarms ? 1
                : 0) };
    }

    /**
     * Creates a read-only DBR_PUT_ACKT. All read operations are delegated to
     * <code>delegate</code>.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessConfigureAcknowledgeTransientAlarmsImpl(
            ChannelAccessConfigureAcknowledgeTransientAlarmsImpl delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     */
    private ChannelAccessConfigureAcknowledgeTransientAlarmsImpl() {
    }

    @Override
    public ChannelAccessValueType getType() {
        return ChannelAccessValueType.DBR_PUT_ACKT;
    }

    @Override
    public int getValueSize() {
        if (delegate != null) {
            return delegate.getValueSize();
        }
        return value.length;
    }

    @Override
    public Short getGenericValueElement(int index) {
        if (delegate != null) {
            return delegate.getGenericValueElement(index);
        }
        return value[index];
    }

    @Override
    public boolean isExplicitlyAcknowledgeTransientAlarms() {
        if (delegate != null) {
            return delegate.isExplicitlyAcknowledgeTransientAlarms();
        }
        if (value.length == 0) {
            return false;
        } else {
            return value[0] != (short) 0;
        }
    }

    @Override
    public void setExplicitlyAcknowledgeTransientAlarms(
            boolean explicitlyAcknowledgeTransientAlarms) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value.length == 1) {
            value[0] = (short) (explicitlyAcknowledgeTransientAlarms ? 1 : 0);
        }
    }

    @Override
    public ShortBuffer getValue() {
        if (delegate != null) {
            return delegate.getValue().asReadOnlyBuffer();
        }
        return ShortBuffer.wrap(value);
    }

    @Override
    public void setValue(short[] value) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value == null) {
            throw new NullPointerException();
        }
        this.value = value;
    }

    @Override
    public void serialize(ByteSink byteSink, int count) {
        if (delegate != null) {
            delegate.serialize(byteSink, count);
            return;
        }
        assert (count >= 0);
        assert (count <= value.length);
        if (count == 0) {
            byteSink.putShort((short) 0);
        } else {
            byteSink.putShortArray(value, 0, count);
        }
    }

    @Override
    public ChannelAccessConfigureAcknowledgeTransientAlarms asReadOnlyValue() {
        if (delegate != null) {
            return this;
        }
        return new ChannelAccessConfigureAcknowledgeTransientAlarmsImpl(this);
    }

    @Override
    public boolean equals(Object obj) {
        if (delegate != null) {
            return delegate.equals(obj);
        }
        obj = extractDelegate(obj);
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been checked.
        ChannelAccessConfigureAcknowledgeTransientAlarmsImpl other = (ChannelAccessConfigureAcknowledgeTransientAlarmsImpl) obj;
        return new EqualsBuilder().append(this.value, other.value).isEquals();
    }

    @Override
    public int hashCode() {
        if (delegate != null) {
            return delegate.hashCode();
        }
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(this.value).toHashCode();
    }

    @Override
    public ChannelAccessConfigureAcknowledgeTransientAlarms clone() {
        if (delegate != null) {
            return delegate.clone();
        }
        ChannelAccessConfigureAcknowledgeTransientAlarmsImpl copy = (ChannelAccessConfigureAcknowledgeTransientAlarmsImpl) super
                .clone();
        copy.value = this.value.clone();
        return copy;
    }

    /**
     * Creates a DBR_PUT_ACKT from its serialized representation.
     * 
     * @param byteSource
     *            byte source to read data from.
     * @param count
     *            number of value elements to read.
     * @return DBR_PUT_ACKT instance created from the serialized data read from
     *         <code>byteSource</code>.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    public static ChannelAccessConfigureAcknowledgeTransientAlarmsImpl deserialize(
            ByteSource byteSource, int count) {
        assert (count >= 0);
        ChannelAccessConfigureAcknowledgeTransientAlarmsImpl value = new ChannelAccessConfigureAcknowledgeTransientAlarmsImpl();
        if (count == 0) {
            byteSource.getShort();
            value.value = ArrayUtils.EMPTY_SHORT_ARRAY;
        } else {
            value.value = byteSource.getShortArray(count);
        }
        return value;
    }

}
