/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.math.BigInteger;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.value.ChannelAccessTimeStamp;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;

/**
 * <p>
 * Time-stamp in the format used by Channel Access. A Channel Access time-stamp
 * has nano-second precision and uses a 32-bit unsigned integer for storing the
 * seconds since January 1st, 1990, 00:00:00 UTC and a 32-bit unsigned integer
 * for the nanoseconds part.
 * </p>
 * 
 * <p>
 * For obvious reasons, the nanoseconds part should only have values in the
 * range of 0 to 999999999. This implementation silently enforces those limits
 * by truncating values that are out of this range.
 * </p>
 * 
 * <p>
 * This type cannot be instantiated directly but should be create using
 * {@link ChannelAccessValueFactory#createTimeStamp(int, int)}.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessTimeStampImpl implements ChannelAccessTimeStamp {

    private static final BigInteger ONE_BILLION = BigInteger
            .valueOf(1000000000L);
    private static final long OFFSET_EPICS_TO_UNIX_EPOCH = 631152000L;

    private int seconds;
    private int nanoseconds;

    /**
     * Creates a time-stamp <code>seconds</code> seconds and
     * <code>nanoseconds</code> nanoseconds after epoch. The EPICS epoch is
     * January 1st, 1990, 00:00:00 UTC. EPICS uses unsigned integers, so a
     * negative value for <code>seconds</code> actually indicates a value
     * greater than the {@link Integer#MAX_VALUE maximum positive integer value}
     * . This constructor should only be used by
     * {@link ChannelAccessValueFactory}.
     * 
     * @param seconds
     *            number of seconds since epoch (interpreted as an unsigned
     *            32-bit integer).
     * @param nanoseconds
     *            number of nanoseconds to be added for higher precision.
     */
    public ChannelAccessTimeStampImpl(int seconds, int nanoseconds) {
        this.seconds = seconds;
        if (nanoseconds > 999999999) {
            nanoseconds = 999999999;
        } else if (nanoseconds < 0) {
            nanoseconds = 0;
        }
        this.nanoseconds = nanoseconds;
    }

    @Override
    public int compareTo(ChannelAccessTimeStamp o) {
        // This solution is not optimal. If someone ever got the stupid idea
        // to implement the ChannelAccessTimeStamp interface in a different
        // class, compareTo might return zero when equals returns false.
        // However, there is no way around this, because we cannot consistently
        // return a positive or negative value for an unknown class.
        if (!(o instanceof ChannelAccessTimeStampImpl)) {
            return new CompareToBuilder().append(this.seconds, o.getSeconds())
                    .append(this.nanoseconds, o.getNanoseconds()).build();
        }
        ChannelAccessTimeStampImpl other = (ChannelAccessTimeStampImpl) o;
        return new CompareToBuilder().append(this.seconds, other.seconds)
                .append(this.nanoseconds, other.nanoseconds).build();
    }

    @Override
    public int getSeconds() {
        return seconds;
    }

    @Override
    public int getNanoseconds() {
        return nanoseconds;
    }

    @Override
    public BigInteger toNanosecondsEPICS() {
        long longSeconds = seconds;
        if (longSeconds < 0) {
            longSeconds += 4294967296L;
        }
        return BigInteger.valueOf(longSeconds).multiply(ONE_BILLION)
                .add(BigInteger.valueOf(nanoseconds));
    }

    @Override
    public BigInteger toNanosecondsUNIX() {
        long longSeconds = seconds;
        if (longSeconds < 0) {
            longSeconds += 4294967296L;
        }
        longSeconds += OFFSET_EPICS_TO_UNIX_EPOCH;
        return BigInteger.valueOf(longSeconds).multiply(ONE_BILLION)
                .add(BigInteger.valueOf(nanoseconds));
    }

    @Override
    public long toJavaTimeStamp() {
        long longSeconds = seconds;
        if (longSeconds < 0) {
            longSeconds += 4294967296L;
        }
        longSeconds += OFFSET_EPICS_TO_UNIX_EPOCH;
        return longSeconds * 1000L;
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().append(seconds).append(nanoseconds)
                .toHashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof ChannelAccessTimeStampImpl)) {
            return false;
        }
        ChannelAccessTimeStampImpl other = (ChannelAccessTimeStampImpl) obj;
        return this.seconds == other.seconds
                && this.nanoseconds == other.nanoseconds;
    }

    @Override
    public String toString() {
        StringBuilder nanosecondsString = new StringBuilder(9);
        nanosecondsString.append(nanoseconds);
        while (nanosecondsString.length() < 9) {
            nanosecondsString.insert(0, '0');
        }
        // We use a generic date format and the US locale, so that the string
        // always looks the same, regardless of the platform settings.
        return new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.######### zzz",
                Locale.US).format(new Date(toJavaTimeStamp())).replace(
                "#########", nanosecondsString);
    }

}
