/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;

import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.junit.Test;

/**
 * Tests for the {@link ChannelAccessValueFactory}. This includes some tests of
 * the values created.
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessValueFactoryTest {

    private final static Charset UTF8 = Charset.forName("UTF8");

    @Test
    public void createTimeStamp() {
        int seconds = 123456789;
        int nanoseconds = 999999999;
        ChannelAccessTimeStamp timeStamp = ChannelAccessValueFactory
                .createTimeStamp(seconds, nanoseconds);
        assertEquals(seconds, timeStamp.getSeconds());
        assertEquals(nanoseconds, timeStamp.getNanoseconds());
    }

    @Test
    public void createTimeStampTooLargeNanoseconds() {
        int seconds = 123456789;
        int nanoseconds = 1000000000;
        ChannelAccessTimeStamp timeStamp = ChannelAccessValueFactory
                .createTimeStamp(seconds, nanoseconds);
        assertEquals(seconds, timeStamp.getSeconds());
        assertEquals(999999999, timeStamp.getNanoseconds());
    }

    @Test
    public void createTimeStampTooSmallNanoseconds() {
        int seconds = 123456789;
        int nanoseconds = -1;
        ChannelAccessTimeStamp timeStamp = ChannelAccessValueFactory
                .createTimeStamp(seconds, nanoseconds);
        assertEquals(seconds, timeStamp.getSeconds());
        assertEquals(0, timeStamp.getNanoseconds());
    }

    @Test
    public void createString() {
        List<String> stringList = Arrays.asList(new String[] { "abc", "def",
                "123456789012345678901234567890123456789012" });
        byte[] rawArray = new byte[] { 97, 98, 99, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 100, 101, 102, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 49, 50, 51, 52, 53, 54, 55, 56, 57, 48,
                49, 50, 51, 52, 53, 54, 55, 56, 57, 48, 49, 50, 51, 52, 53, 54,
                55, 56, 57, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 0 };
        ChannelAccessString string1 = ChannelAccessValueFactory.createString(
                stringList, UTF8);
        assertEquals(ChannelAccessValueType.DBR_STRING, string1.getType());
        assertEquals(3, string1.getValueSize());
        assertEquals(stringList, string1.getValue());
        assertArrayEquals(rawArray, string1.getRawValue());
        assertEquals(string1.getGenericValueElement(1), stringList.get(1));
        ChannelAccessString string2 = (ChannelAccessString) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_STRING, UTF8);
        string2.setRawValue(rawArray);
        assertNotEquals(string1, string2);
        string2.setValue(stringList);
        assertEquals(string1, string2);
        assertEquals(string1.hashCode(), string2.hashCode());
    }

    @Test
    public void createAlarmString() {
        List<String> stringList = Arrays.asList(new String[] { "abc", "def",
                "123456789012345678901234567890123456789012" });
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.LINK;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.INVALID_ALARM;
        ChannelAccessAlarmString string1 = ChannelAccessValueFactory
                .createAlarmString(stringList, alarmStatus, alarmSeverity, UTF8);
        assertEquals(ChannelAccessValueType.DBR_STS_STRING, string1.getType());
        assertEquals(stringList, string1.getValue());
        assertEquals(alarmStatus, string1.getAlarmStatus());
        assertEquals(alarmSeverity, string1.getAlarmSeverity());
        assertEquals(string1.getGenericValueElement(1), stringList.get(1));
        ChannelAccessAlarmString string2 = (ChannelAccessAlarmString) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_STS_STRING, UTF8);
        string2.getValue().addAll(stringList);
        assertNotEquals(string1, string2);
        string2.setAlarmStatus(alarmStatus);
        string2.setAlarmSeverity(alarmSeverity);
        assertEquals(string1, string2);
        assertEquals(string1.hashCode(), string2.hashCode());
    }

    @Test
    public void createTimeString() {
        List<String> stringList = Collections.emptyList();
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.HIGH;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        int seconds = -10000;
        int nanoseconds = 50;
        ChannelAccessTimeString string1 = ChannelAccessValueFactory
                .createTimeString(stringList, alarmStatus, alarmSeverity,
                        seconds, nanoseconds, UTF8);
        assertEquals(ChannelAccessValueType.DBR_TIME_STRING, string1.getType());
        ChannelAccessTimeString string2 = (ChannelAccessTimeString) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_TIME_STRING, UTF8);
        assertNotEquals(string1, string2);
        string2.setAlarmStatus(alarmStatus);
        string2.setAlarmSeverity(alarmSeverity);
        string2.setTimeStamp(string1.getTimeStamp());
        assertEquals(string1, string2);
        assertEquals(string1.hashCode(), string2.hashCode());
    }

    @Test
    public void createGraphicsString() {
        List<String> stringList = Arrays.asList(new String[] { "abc", "def",
                "123456789012345678901234567890123456789012" });
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.LINK;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.INVALID_ALARM;
        ChannelAccessAlarmString string1 = ChannelAccessValueFactory
                .createGraphicsString(stringList, alarmStatus, alarmSeverity,
                        UTF8);
        assertEquals(ChannelAccessValueType.DBR_GR_STRING, string1.getType());
        ChannelAccessAlarmString string2 = (ChannelAccessAlarmString) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_GR_STRING, UTF8);
        assertNotEquals(string1, string2);
        string2.setValue(stringList);
        string2.setAlarmStatus(alarmStatus);
        string2.setAlarmSeverity(alarmSeverity);
        assertEquals(string1, string2);
        assertEquals(string1.hashCode(), string2.hashCode());
    }

    @Test
    public void createControlsString() {
        List<String> stringList = Arrays.asList(new String[] { "abc", "def",
                "123456789012345678901234567890123456789012" });
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.LINK;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.INVALID_ALARM;
        ChannelAccessAlarmString string1 = ChannelAccessValueFactory
                .createControlsString(stringList, alarmStatus, alarmSeverity,
                        UTF8);
        assertEquals(ChannelAccessValueType.DBR_CTRL_STRING, string1.getType());
        ChannelAccessAlarmString string2 = (ChannelAccessAlarmString) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_CTRL_STRING, UTF8);
        assertNotEquals(string1, string2);
        string2.setValue(stringList);
        string2.setAlarmStatus(alarmStatus);
        string2.setAlarmSeverity(alarmSeverity);
        assertEquals(string1, string2);
        assertEquals(string1.hashCode(), string2.hashCode());
    }

    @Test
    public void createShort() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessShort short1 = ChannelAccessValueFactory
                .createShort(shortArray);
        assertEquals(ChannelAccessValueType.DBR_SHORT, short1.getType());
        assertEquals(short1.getValueSize(), 4);
        assertArrayEquals(shortArray, short1.getValue().array());
        assertEquals(shortArray[0], short1.getGenericValueElement(0)
                .shortValue());
        ChannelAccessShort short2 = (ChannelAccessShort) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_SHORT, UTF8);
        assertNotEquals(short1, short2);
        short2.setValue(shortArray);
        assertEquals(short1, short2);
        assertEquals(short1.hashCode(), short2.hashCode());
    }

    @Test
    public void createAlarmShort() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        ChannelAccessAlarmShort short1 = ChannelAccessValueFactory
                .createAlarmShort(shortArray, alarmStatus, alarmSeverity);
        assertEquals(ChannelAccessValueType.DBR_STS_SHORT, short1.getType());
        assertEquals(alarmStatus, short1.getAlarmStatus());
        assertEquals(alarmSeverity, short1.getAlarmSeverity());
        ChannelAccessAlarmShort short2 = (ChannelAccessAlarmShort) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_STS_SHORT, UTF8);
        assertNotEquals(short1, short2);
        short2.setValue(shortArray);
        assertNotEquals(short1, short2);
        short2.setAlarmStatus(alarmStatus);
        assertNotEquals(short1, short2);
        short2.setAlarmSeverity(alarmSeverity);
        assertEquals(short1, short2);
        assertEquals(short1.hashCode(), short2.hashCode());
    }

    @Test
    public void createTimeShort() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        int seconds = 123;
        int nanoseconds = 456;
        ChannelAccessTimeShort short1 = ChannelAccessValueFactory
                .createTimeShort(shortArray, alarmStatus, alarmSeverity,
                        seconds, nanoseconds);
        assertEquals(ChannelAccessValueType.DBR_TIME_SHORT, short1.getType());
        assertEquals(seconds, short1.getTimeSeconds());
        assertEquals(nanoseconds, short1.getTimeNanoseconds());
        ChannelAccessTimeShort short2 = (ChannelAccessTimeShort) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_TIME_SHORT, UTF8);
        assertNotEquals(short1, short2);
        short2.setValue(shortArray);
        assertNotEquals(short1, short2);
        short2.setAlarmStatus(alarmStatus);
        assertNotEquals(short1, short2);
        short2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(short1, short2);
        short2.setTimeSeconds(seconds);
        assertNotEquals(short1, short2);
        short2.setTimeNanoseconds(nanoseconds);
        assertEquals(short1, short2);
        assertEquals(short1.hashCode(), short2.hashCode());
    }

    @Test
    public void createGraphicsShort() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        String engineeringUnits = "abc";
        byte[] rawEngineeringUnits = new byte[] { 97, 98, 99, 0, 0, 0, 0, 0 };
        short upperDisplayLimit = 1000;
        short lowerDisplayLimit = -1000;
        short upperAlarmLimit = 456;
        short upperWarningLimit = 123;
        short lowerWarningLimit = 50;
        short lowerAlarmLimit = 1;
        ChannelAccessGraphicsShort short1 = ChannelAccessValueFactory
                .createGraphicsShort(shortArray, alarmStatus, alarmSeverity,
                        engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                        upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                        lowerAlarmLimit, UTF8);
        assertEquals(ChannelAccessValueType.DBR_GR_SHORT, short1.getType());
        assertEquals(engineeringUnits, short1.getUnits());
        assertArrayEquals(rawEngineeringUnits, short1.getRawUnits());
        assertEquals(upperDisplayLimit, short1.getUpperDisplayLimit());
        assertEquals(lowerDisplayLimit, short1.getLowerDisplayLimit());
        assertEquals(upperAlarmLimit, short1.getUpperAlarmLimit());
        assertEquals(upperWarningLimit, short1.getUpperWarningLimit());
        assertEquals(lowerWarningLimit, short1.getLowerWarningLimit());
        assertEquals(lowerAlarmLimit, short1.getLowerAlarmLimit());
        ChannelAccessGraphicsShort short2 = (ChannelAccessGraphicsShort) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_GR_SHORT, UTF8);
        assertNotEquals(short1, short2);
        short2.setValue(shortArray);
        short2.setAlarmStatus(alarmStatus);
        short2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(short1, short2);
        short2.setRawUnits(rawEngineeringUnits);
        assertNotEquals(short1, short2);
        short2.setUpperDisplayLimit(upperDisplayLimit);
        assertNotEquals(short1, short2);
        short2.setLowerDisplayLimit(lowerDisplayLimit);
        assertNotEquals(short1, short2);
        short2.setUpperAlarmLimit(upperAlarmLimit);
        assertNotEquals(short1, short2);
        short2.setUpperWarningLimit(upperWarningLimit);
        assertNotEquals(short1, short2);
        short2.setLowerWarningLimit(lowerWarningLimit);
        assertNotEquals(short1, short2);
        short2.setLowerAlarmLimit(lowerAlarmLimit);
        assertEquals(short1, short2);
        assertEquals(short1.hashCode(), short2.hashCode());
    }

    @Test
    public void createControlsShort() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        String engineeringUnits = "abc";
        short upperDisplayLimit = 1000;
        short lowerDisplayLimit = -1000;
        short upperAlarmLimit = 456;
        short upperWarningLimit = 123;
        short lowerWarningLimit = 50;
        short lowerAlarmLimit = 1;
        short upperControlLimit = 8;
        short lowerControlLimit = -25;
        ChannelAccessControlsShort short1 = ChannelAccessValueFactory
                .createControlsShort(shortArray, alarmStatus, alarmSeverity,
                        engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                        upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                        lowerAlarmLimit, upperControlLimit, lowerControlLimit,
                        UTF8);
        assertEquals(ChannelAccessValueType.DBR_CTRL_SHORT, short1.getType());
        assertEquals(upperControlLimit, short1.getUpperControlLimit());
        assertEquals(lowerControlLimit, short1.getLowerControlLimit());
        ChannelAccessControlsShort short2 = (ChannelAccessControlsShort) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_CTRL_SHORT, UTF8);
        assertNotEquals(short1, short2);
        short2.setValue(shortArray);
        short2.setAlarmStatus(alarmStatus);
        short2.setAlarmSeverity(alarmSeverity);
        short2.setUnits(engineeringUnits);
        short2.setUpperDisplayLimit(upperDisplayLimit);
        short2.setLowerDisplayLimit(lowerDisplayLimit);
        short2.setUpperAlarmLimit(upperAlarmLimit);
        short2.setUpperWarningLimit(upperWarningLimit);
        short2.setLowerWarningLimit(lowerWarningLimit);
        short2.setLowerAlarmLimit(lowerAlarmLimit);
        assertNotEquals(short1, short2);
        short2.setUpperControlLimit(upperControlLimit);
        assertNotEquals(short1, short2);
        short2.setLowerControlLimit(lowerControlLimit);
        assertEquals(short1, short2);
        assertEquals(short1.hashCode(), short2.hashCode());
    }

    @Test
    public void createFloat() {
        float[] floatArray = new float[] { 5, -7, 2174, -35 };
        ChannelAccessFloat float1 = ChannelAccessValueFactory
                .createFloat(floatArray);
        assertEquals(ChannelAccessValueType.DBR_FLOAT, float1.getType());
        assertEquals(float1.getValueSize(), 4);
        assertArrayEquals(floatArray, float1.getValue().array(), 0.0f);
        assertEquals(floatArray[0], float1.getGenericValueElement(0)
                .floatValue(), 0.0f);
        ChannelAccessFloat float2 = (ChannelAccessFloat) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_FLOAT, UTF8);
        assertNotEquals(float1, float2);
        float2.setValue(floatArray);
        assertEquals(float1, float2);
        assertEquals(float1.hashCode(), float2.hashCode());
    }

    @Test
    public void createAlarmFloat() {
        float[] floatArray = new float[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.HIGH;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.NO_ALARM;
        ChannelAccessAlarmFloat float1 = ChannelAccessValueFactory
                .createAlarmFloat(floatArray, alarmStatus, alarmSeverity);
        assertEquals(ChannelAccessValueType.DBR_STS_FLOAT, float1.getType());
        assertEquals(alarmStatus, float1.getAlarmStatus());
        assertEquals(alarmSeverity, float1.getAlarmSeverity());
        ChannelAccessAlarmFloat float2 = (ChannelAccessAlarmFloat) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_STS_FLOAT, UTF8);
        assertNotEquals(float1, float2);
        float2.setValue(floatArray);
        assertNotEquals(float1, float2);
        float2.setAlarmStatus(alarmStatus);
        float2.setAlarmSeverity(alarmSeverity);
        assertEquals(float1, float2);
        assertEquals(float1.hashCode(), float2.hashCode());
    }

    @Test
    public void createTimeFloat() {
        float[] floatArray = new float[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        int seconds = 123;
        int nanoseconds = 456;
        ChannelAccessTimeFloat float1 = ChannelAccessValueFactory
                .createTimeFloat(floatArray, alarmStatus, alarmSeverity,
                        seconds, nanoseconds);
        assertEquals(ChannelAccessValueType.DBR_TIME_FLOAT, float1.getType());
        assertEquals(seconds, float1.getTimeSeconds());
        assertEquals(nanoseconds, float1.getTimeNanoseconds());
        ChannelAccessTimeFloat float2 = (ChannelAccessTimeFloat) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_TIME_FLOAT, UTF8);
        assertNotEquals(float1, float2);
        float2.setValue(floatArray);
        assertNotEquals(float1, float2);
        float2.setAlarmStatus(alarmStatus);
        assertNotEquals(float1, float2);
        float2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(float1, float2);
        float2.setTimeSeconds(seconds);
        assertNotEquals(float1, float2);
        float2.setTimeNanoseconds(nanoseconds);
        assertEquals(float1, float2);
        assertEquals(float1.hashCode(), float2.hashCode());
    }

    @Test
    public void createGraphicsFloat() {
        float[] floatArray = new float[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        short precision = 3;
        String engineeringUnits = "abcdefgh";
        byte[] rawEngineeringUnits = new byte[] { 97, 98, 99, 100, 101, 102,
                103, 0 };
        float upperDisplayLimit = 1000.0f;
        float lowerDisplayLimit = -1000.0f;
        float upperAlarmLimit = 456.0f;
        float upperWarningLimit = 123.0f;
        float lowerWarningLimit = 50.0f;
        float lowerAlarmLimit = 1.0f;
        ChannelAccessGraphicsFloat float1 = ChannelAccessValueFactory
                .createGraphicsFloat(floatArray, alarmStatus, alarmSeverity,
                        precision, engineeringUnits, upperDisplayLimit,
                        lowerDisplayLimit, upperAlarmLimit, upperWarningLimit,
                        lowerWarningLimit, lowerAlarmLimit, UTF8);
        assertEquals(ChannelAccessValueType.DBR_GR_FLOAT, float1.getType());
        assertEquals(precision, float1.getPrecision());
        assertEquals(engineeringUnits, float1.getUnits());
        assertArrayEquals(rawEngineeringUnits, float1.getRawUnits());
        assertEquals(upperDisplayLimit, float1.getUpperDisplayLimit(), 0.0f);
        assertEquals(lowerDisplayLimit, float1.getLowerDisplayLimit(), 0.0f);
        assertEquals(upperAlarmLimit, float1.getUpperAlarmLimit(), 0.0f);
        assertEquals(upperWarningLimit, float1.getUpperWarningLimit(), 0.0f);
        assertEquals(lowerWarningLimit, float1.getLowerWarningLimit(), 0.0f);
        assertEquals(lowerAlarmLimit, float1.getLowerAlarmLimit(), 0.0f);
        ChannelAccessGraphicsFloat float2 = (ChannelAccessGraphicsFloat) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_GR_FLOAT, UTF8);
        assertNotEquals(float1, float2);
        float2.setValue(floatArray);
        float2.setAlarmStatus(alarmStatus);
        float2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(float1, float2);
        float2.setPrecision(precision);
        assertNotEquals(float1, float2);
        float2.setUpperDisplayLimit(upperDisplayLimit);
        assertNotEquals(float1, float2);
        float2.setLowerDisplayLimit(lowerDisplayLimit);
        assertNotEquals(float1, float2);
        float2.setUpperAlarmLimit(upperAlarmLimit);
        assertNotEquals(float1, float2);
        float2.setUpperWarningLimit(upperWarningLimit);
        assertNotEquals(float1, float2);
        float2.setLowerWarningLimit(lowerWarningLimit);
        assertNotEquals(float1, float2);
        float2.setLowerAlarmLimit(lowerAlarmLimit);
        assertNotEquals(float1, float2);
        float2.setRawUnits(rawEngineeringUnits);
        assertNotEquals(float1, float2);
        float2.setUnits(engineeringUnits);
        assertEquals(float1, float2);
        assertEquals(float1.hashCode(), float2.hashCode());
    }

    @Test
    public void createControlsFloat() {
        float[] floatArray = new float[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        short precision = 3;
        String engineeringUnits = "abcdefgh";
        float upperDisplayLimit = 1000.0f;
        float lowerDisplayLimit = -1000.0f;
        float upperAlarmLimit = 456.0f;
        float upperWarningLimit = 123.0f;
        float lowerWarningLimit = 50.0f;
        float lowerAlarmLimit = 1.0f;
        float upperControlLimit = Float.POSITIVE_INFINITY;
        float lowerControlLimit = Float.NaN;
        ChannelAccessControlsFloat float1 = ChannelAccessValueFactory
                .createControlsFloat(floatArray, alarmStatus, alarmSeverity,
                        precision, engineeringUnits, upperDisplayLimit,
                        lowerDisplayLimit, upperAlarmLimit, upperWarningLimit,
                        lowerWarningLimit, lowerAlarmLimit, upperControlLimit,
                        lowerControlLimit, UTF8);
        assertEquals(ChannelAccessValueType.DBR_CTRL_FLOAT, float1.getType());
        assertEquals(upperControlLimit, float1.getUpperControlLimit(), 0.0f);
        assertEquals(lowerControlLimit, float1.getLowerControlLimit(), 0.0f);
        ChannelAccessControlsFloat float2 = (ChannelAccessControlsFloat) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_CTRL_FLOAT, UTF8);
        assertNotEquals(float1, float2);
        float2.setValue(floatArray);
        float2.setAlarmStatus(alarmStatus);
        float2.setAlarmSeverity(alarmSeverity);
        float2.setPrecision(precision);
        float2.setUnits(engineeringUnits);
        float2.setUpperDisplayLimit(upperDisplayLimit);
        float2.setLowerDisplayLimit(lowerDisplayLimit);
        float2.setUpperAlarmLimit(upperAlarmLimit);
        float2.setUpperWarningLimit(upperWarningLimit);
        float2.setLowerWarningLimit(lowerWarningLimit);
        float2.setLowerAlarmLimit(lowerAlarmLimit);
        assertNotEquals(float1, float2);
        float2.setUpperControlLimit(upperControlLimit);
        assertNotEquals(float1, float2);
        float2.setLowerControlLimit(lowerControlLimit);
        assertEquals(float1, float2);
        assertEquals(float1.hashCode(), float2.hashCode());
    }

    @Test
    public void createEnum() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessEnum enum1 = ChannelAccessValueFactory
                .createEnum(shortArray);
        assertEquals(ChannelAccessValueType.DBR_ENUM, enum1.getType());
        assertEquals(enum1.getValueSize(), 4);
        assertArrayEquals(shortArray, enum1.getValue().array());
        assertEquals(shortArray[0], enum1.getGenericValueElement(0)
                .shortValue());
        ChannelAccessEnum enum2 = (ChannelAccessEnum) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_ENUM, UTF8);
        assertNotEquals(enum1, enum2);
        enum2.setValue(shortArray);
        assertEquals(enum1, enum2);
        assertEquals(enum1.hashCode(), enum2.hashCode());
    }

    @Test
    public void createAlarmEnum() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        ChannelAccessAlarmEnum enum1 = ChannelAccessValueFactory
                .createAlarmEnum(shortArray, alarmStatus, alarmSeverity);
        assertEquals(ChannelAccessValueType.DBR_STS_ENUM, enum1.getType());
        assertEquals(alarmStatus, enum1.getAlarmStatus());
        assertEquals(alarmSeverity, enum1.getAlarmSeverity());
        ChannelAccessAlarmEnum enum2 = (ChannelAccessAlarmEnum) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_STS_ENUM, UTF8);
        assertNotEquals(enum1, enum2);
        enum2.setValue(shortArray);
        assertNotEquals(enum1, enum2);
        enum2.setAlarmStatus(alarmStatus);
        assertNotEquals(enum1, enum2);
        enum2.setAlarmSeverity(alarmSeverity);
        assertEquals(enum1, enum2);
        assertEquals(enum1.hashCode(), enum2.hashCode());
    }

    @Test
    public void createTimeEnum() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        int seconds = 123;
        int nanoseconds = 456;
        ChannelAccessTimeEnum enum1 = ChannelAccessValueFactory.createTimeEnum(
                shortArray, alarmStatus, alarmSeverity, seconds, nanoseconds);
        assertEquals(ChannelAccessValueType.DBR_TIME_ENUM, enum1.getType());
        assertEquals(seconds, enum1.getTimeSeconds());
        assertEquals(nanoseconds, enum1.getTimeNanoseconds());
        ChannelAccessTimeEnum enum2 = (ChannelAccessTimeEnum) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_TIME_ENUM, UTF8);
        assertNotEquals(enum1, enum2);
        enum2.setValue(shortArray);
        assertNotEquals(enum1, enum2);
        enum2.setAlarmStatus(alarmStatus);
        assertNotEquals(enum1, enum2);
        enum2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(enum1, enum2);
        enum2.setTimeSeconds(seconds);
        assertNotEquals(enum1, enum2);
        enum2.setTimeNanoseconds(nanoseconds);
        assertEquals(enum1, enum2);
        assertEquals(enum1.hashCode(), enum2.hashCode());
    }

    @Test
    public void createGraphicsEnum() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        List<String> labels = Arrays.asList(new String[] { "", "abc", "def" });
        byte[] rawLabels = new byte[416];
        rawLabels[26] = 97;
        rawLabels[27] = 98;
        rawLabels[28] = 99;
        rawLabels[52] = 100;
        rawLabels[53] = 101;
        rawLabels[54] = 102;
        ChannelAccessGraphicsEnum enum1 = ChannelAccessValueFactory
                .createGraphicsEnum(shortArray, alarmStatus, alarmSeverity,
                        labels, UTF8);
        assertEquals(ChannelAccessValueType.DBR_GR_ENUM, enum1.getType());
        assertEquals(labels, enum1.getLabels());
        assertArrayEquals(rawLabels, enum1.getRawLabels());
        ChannelAccessGraphicsEnum enum2 = (ChannelAccessGraphicsEnum) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_GR_ENUM, UTF8);
        assertNotEquals(enum1, enum2);
        enum2.setValue(shortArray);
        enum2.setAlarmStatus(alarmStatus);
        enum2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(enum1, enum2);
        enum2.setRawLabels(rawLabels, 3);
        assertEquals(enum1, enum2);
        assertEquals(enum1.hashCode(), enum2.hashCode());
    }

    @Test
    public void createControlsEnum() {
        short[] shortArray = new short[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        List<String> labels = Arrays.asList(new String[] { "abc", "def" });
        ChannelAccessGraphicsEnum enum1 = ChannelAccessValueFactory
                .createControlsEnum(shortArray, alarmStatus, alarmSeverity,
                        labels, UTF8);
        assertEquals(ChannelAccessValueType.DBR_CTRL_ENUM, enum1.getType());
        ChannelAccessGraphicsEnum enum2 = (ChannelAccessGraphicsEnum) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_CTRL_ENUM, UTF8);
        assertNotEquals(enum1, enum2);
        enum2.setAlarmStatus(alarmStatus);
        enum2.setAlarmSeverity(alarmSeverity);
        enum2.setValue(shortArray);
        enum2.getLabels().addAll(labels);
        assertEquals(enum1, enum2);
        assertEquals(enum1.hashCode(), enum2.hashCode());
    }

    @Test
    public void createChar() {
        byte[] byteArray = new byte[] { 5, -7, 116, -35 };
        ChannelAccessChar char1 = ChannelAccessValueFactory
                .createChar(byteArray);
        assertEquals(ChannelAccessValueType.DBR_CHAR, char1.getType());
        assertEquals(char1.getValueSize(), 4);
        assertArrayEquals(byteArray, char1.getValue().array());
        assertEquals(byteArray[0], char1.getGenericValueElement(0).byteValue());
        ChannelAccessChar char2 = (ChannelAccessChar) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_CHAR, UTF8);
        assertNotEquals(char1, char2);
        char2.setValue(byteArray);
        assertEquals(char1, char2);
        assertEquals(char1.hashCode(), char2.hashCode());
    }

    @Test
    public void createAlarmChar() {
        byte[] byteArray = new byte[] { 5, -7, 116, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        ChannelAccessAlarmChar char1 = ChannelAccessValueFactory
                .createAlarmChar(byteArray, alarmStatus, alarmSeverity);
        assertEquals(ChannelAccessValueType.DBR_STS_CHAR, char1.getType());
        assertEquals(alarmStatus, char1.getAlarmStatus());
        assertEquals(alarmSeverity, char1.getAlarmSeverity());
        ChannelAccessAlarmChar char2 = (ChannelAccessAlarmChar) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_STS_CHAR, UTF8);
        assertNotEquals(char1, char2);
        char2.setValue(byteArray);
        assertNotEquals(char1, char2);
        char2.setAlarmStatus(alarmStatus);
        assertNotEquals(char1, char2);
        char2.setAlarmSeverity(alarmSeverity);
        assertEquals(char1, char2);
        assertEquals(char1.hashCode(), char2.hashCode());
    }

    @Test
    public void createTimeChar() {
        byte[] byteArray = new byte[] { 5, -7, 116, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        int seconds = 123;
        int nanoseconds = 456;
        ChannelAccessTimeChar char1 = ChannelAccessValueFactory.createTimeChar(
                byteArray, alarmStatus, alarmSeverity, seconds, nanoseconds);
        assertEquals(ChannelAccessValueType.DBR_TIME_CHAR, char1.getType());
        assertEquals(seconds, char1.getTimeSeconds());
        assertEquals(nanoseconds, char1.getTimeNanoseconds());
        ChannelAccessTimeChar char2 = (ChannelAccessTimeChar) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_TIME_CHAR, UTF8);
        assertNotEquals(char1, char2);
        char2.setValue(byteArray);
        assertNotEquals(char1, char2);
        char2.setAlarmStatus(alarmStatus);
        assertNotEquals(char1, char2);
        char2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(char1, char2);
        char2.setTimeSeconds(seconds);
        assertNotEquals(char1, char2);
        char2.setTimeNanoseconds(nanoseconds);
        assertEquals(char1, char2);
        assertEquals(char1.hashCode(), char2.hashCode());
    }

    @Test
    public void createGraphicsChar() {
        byte[] byteArray = new byte[] { 5, -7, 116, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        String engineeringUnits = "abc";
        byte[] rawEngineeringUnits = new byte[] { 97, 98, 99, 0, 0, 0, 0, 0 };
        byte upperDisplayLimit = 100;
        byte lowerDisplayLimit = -100;
        byte upperAlarmLimit = 45;
        byte upperWarningLimit = 123;
        byte lowerWarningLimit = 50;
        byte lowerAlarmLimit = 1;
        ChannelAccessGraphicsChar char1 = ChannelAccessValueFactory
                .createGraphicsChar(byteArray, alarmStatus, alarmSeverity,
                        engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                        upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                        lowerAlarmLimit, UTF8);
        assertEquals(ChannelAccessValueType.DBR_GR_CHAR, char1.getType());
        assertEquals(engineeringUnits, char1.getUnits());
        assertArrayEquals(rawEngineeringUnits, char1.getRawUnits());
        assertEquals(upperDisplayLimit, char1.getUpperDisplayLimit());
        assertEquals(lowerDisplayLimit, char1.getLowerDisplayLimit());
        assertEquals(upperAlarmLimit, char1.getUpperAlarmLimit());
        assertEquals(upperWarningLimit, char1.getUpperWarningLimit());
        assertEquals(lowerWarningLimit, char1.getLowerWarningLimit());
        assertEquals(lowerAlarmLimit, char1.getLowerAlarmLimit());
        ChannelAccessGraphicsChar char2 = (ChannelAccessGraphicsChar) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_GR_CHAR, UTF8);
        assertNotEquals(char1, char2);
        char2.setValue(byteArray);
        char2.setAlarmStatus(alarmStatus);
        char2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(char1, char2);
        char2.setRawUnits(rawEngineeringUnits);
        assertNotEquals(char1, char2);
        char2.setUpperDisplayLimit(upperDisplayLimit);
        assertNotEquals(char1, char2);
        char2.setLowerDisplayLimit(lowerDisplayLimit);
        assertNotEquals(char1, char2);
        char2.setUpperAlarmLimit(upperAlarmLimit);
        assertNotEquals(char1, char2);
        char2.setUpperWarningLimit(upperWarningLimit);
        assertNotEquals(char1, char2);
        char2.setLowerWarningLimit(lowerWarningLimit);
        assertNotEquals(char1, char2);
        char2.setLowerAlarmLimit(lowerAlarmLimit);
        assertEquals(char1, char2);
        assertEquals(char1.hashCode(), char2.hashCode());
    }

    @Test
    public void createControlsChar() {
        byte[] byteArray = new byte[] { 5, -7, 116, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        String engineeringUnits = "abc";
        byte upperDisplayLimit = 100;
        byte lowerDisplayLimit = -100;
        byte upperAlarmLimit = 45;
        byte upperWarningLimit = 123;
        byte lowerWarningLimit = 50;
        byte lowerAlarmLimit = 1;
        byte upperControlLimit = 8;
        byte lowerControlLimit = -25;
        ChannelAccessControlsChar char1 = ChannelAccessValueFactory
                .createControlsChar(byteArray, alarmStatus, alarmSeverity,
                        engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                        upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                        lowerAlarmLimit, upperControlLimit, lowerControlLimit,
                        UTF8);
        assertEquals(ChannelAccessValueType.DBR_CTRL_CHAR, char1.getType());
        assertEquals(upperControlLimit, char1.getUpperControlLimit());
        assertEquals(lowerControlLimit, char1.getLowerControlLimit());
        ChannelAccessControlsChar char2 = (ChannelAccessControlsChar) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_CTRL_CHAR, UTF8);
        assertNotEquals(char1, char2);
        char2.setValue(byteArray);
        char2.setAlarmStatus(alarmStatus);
        char2.setAlarmSeverity(alarmSeverity);
        char2.setUnits(engineeringUnits);
        char2.setUpperDisplayLimit(upperDisplayLimit);
        char2.setLowerDisplayLimit(lowerDisplayLimit);
        char2.setUpperAlarmLimit(upperAlarmLimit);
        char2.setUpperWarningLimit(upperWarningLimit);
        char2.setLowerWarningLimit(lowerWarningLimit);
        char2.setLowerAlarmLimit(lowerAlarmLimit);
        assertNotEquals(char1, char2);
        char2.setUpperControlLimit(upperControlLimit);
        assertNotEquals(char1, char2);
        char2.setLowerControlLimit(lowerControlLimit);
        assertEquals(char1, char2);
        assertEquals(char1.hashCode(), char2.hashCode());
    }

    @Test
    public void createLong() {
        int[] intArray = new int[] { 5, -7, 116, -35 };
        ChannelAccessLong long1 = ChannelAccessValueFactory
                .createLong(intArray);
        assertEquals(ChannelAccessValueType.DBR_LONG, long1.getType());
        assertEquals(long1.getValueSize(), 4);
        assertArrayEquals(intArray, long1.getValue().array());
        assertEquals(intArray[0], long1.getGenericValueElement(0).intValue());
        ChannelAccessLong long2 = (ChannelAccessLong) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_LONG, UTF8);
        assertNotEquals(long1, long2);
        long2.setValue(intArray);
        assertEquals(long1, long2);
        assertEquals(long1.hashCode(), long2.hashCode());
    }

    @Test
    public void createAlarmLong() {
        int[] intArray = new int[] { 5, -7, 116, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        ChannelAccessAlarmLong long1 = ChannelAccessValueFactory
                .createAlarmLong(intArray, alarmStatus, alarmSeverity);
        assertEquals(ChannelAccessValueType.DBR_STS_LONG, long1.getType());
        assertEquals(alarmStatus, long1.getAlarmStatus());
        assertEquals(alarmSeverity, long1.getAlarmSeverity());
        ChannelAccessAlarmLong long2 = (ChannelAccessAlarmLong) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_STS_LONG, UTF8);
        assertNotEquals(long1, long2);
        long2.setValue(intArray);
        assertNotEquals(long1, long2);
        long2.setAlarmStatus(alarmStatus);
        assertNotEquals(long1, long2);
        long2.setAlarmSeverity(alarmSeverity);
        assertEquals(long1, long2);
        assertEquals(long1.hashCode(), long2.hashCode());
    }

    @Test
    public void createTimeLong() {
        int[] intArray = new int[] { 5, -7, 116, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        int seconds = 123;
        int nanoseconds = 456;
        ChannelAccessTimeLong long1 = ChannelAccessValueFactory.createTimeLong(
                intArray, alarmStatus, alarmSeverity, seconds, nanoseconds);
        assertEquals(ChannelAccessValueType.DBR_TIME_LONG, long1.getType());
        assertEquals(seconds, long1.getTimeSeconds());
        assertEquals(nanoseconds, long1.getTimeNanoseconds());
        ChannelAccessTimeLong long2 = (ChannelAccessTimeLong) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_TIME_LONG, UTF8);
        assertNotEquals(long1, long2);
        long2.setValue(intArray);
        assertNotEquals(long1, long2);
        long2.setAlarmStatus(alarmStatus);
        assertNotEquals(long1, long2);
        long2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(long1, long2);
        long2.setTimeSeconds(seconds);
        assertNotEquals(long1, long2);
        long2.setTimeNanoseconds(nanoseconds);
        assertEquals(long1, long2);
        assertEquals(long1.hashCode(), long2.hashCode());
    }

    @Test
    public void createGraphicsLong() {
        int[] intArray = new int[] { 5, -7, 116, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        String engineeringUnits = "abc";
        byte[] rawEngineeringUnits = new byte[] { 97, 98, 99, 0, 0, 0, 0, 0 };
        int upperDisplayLimit = 100;
        int lowerDisplayLimit = -100;
        int upperAlarmLimit = 45;
        int upperWarningLimit = 123;
        int lowerWarningLimit = 50;
        int lowerAlarmLimit = 1;
        ChannelAccessGraphicsLong long1 = ChannelAccessValueFactory
                .createGraphicsLong(intArray, alarmStatus, alarmSeverity,
                        engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                        upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                        lowerAlarmLimit, UTF8);
        assertEquals(ChannelAccessValueType.DBR_GR_LONG, long1.getType());
        assertEquals(engineeringUnits, long1.getUnits());
        assertArrayEquals(rawEngineeringUnits, long1.getRawUnits());
        assertEquals(upperDisplayLimit, long1.getUpperDisplayLimit());
        assertEquals(lowerDisplayLimit, long1.getLowerDisplayLimit());
        assertEquals(upperAlarmLimit, long1.getUpperAlarmLimit());
        assertEquals(upperWarningLimit, long1.getUpperWarningLimit());
        assertEquals(lowerWarningLimit, long1.getLowerWarningLimit());
        assertEquals(lowerAlarmLimit, long1.getLowerAlarmLimit());
        ChannelAccessGraphicsLong long2 = (ChannelAccessGraphicsLong) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_GR_LONG, UTF8);
        assertNotEquals(long1, long2);
        long2.setValue(intArray);
        long2.setAlarmStatus(alarmStatus);
        long2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(long1, long2);
        long2.setRawUnits(rawEngineeringUnits);
        assertNotEquals(long1, long2);
        long2.setUpperDisplayLimit(upperDisplayLimit);
        assertNotEquals(long1, long2);
        long2.setLowerDisplayLimit(lowerDisplayLimit);
        assertNotEquals(long1, long2);
        long2.setUpperAlarmLimit(upperAlarmLimit);
        assertNotEquals(long1, long2);
        long2.setUpperWarningLimit(upperWarningLimit);
        assertNotEquals(long1, long2);
        long2.setLowerWarningLimit(lowerWarningLimit);
        assertNotEquals(long1, long2);
        long2.setLowerAlarmLimit(lowerAlarmLimit);
        assertEquals(long1, long2);
        assertEquals(long1.hashCode(), long2.hashCode());
    }

    @Test
    public void createControlsLong() {
        int[] intArray = new int[] { 5, -7, 116, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        String engineeringUnits = "abc";
        int upperDisplayLimit = 100;
        int lowerDisplayLimit = -100;
        int upperAlarmLimit = 45;
        int upperWarningLimit = 123;
        int lowerWarningLimit = 50;
        int lowerAlarmLimit = 1;
        int upperControlLimit = 8;
        int lowerControlLimit = -25;
        ChannelAccessControlsLong long1 = ChannelAccessValueFactory
                .createControlsLong(intArray, alarmStatus, alarmSeverity,
                        engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                        upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                        lowerAlarmLimit, upperControlLimit, lowerControlLimit,
                        UTF8);
        assertEquals(ChannelAccessValueType.DBR_CTRL_LONG, long1.getType());
        assertEquals(upperControlLimit, long1.getUpperControlLimit());
        assertEquals(lowerControlLimit, long1.getLowerControlLimit());
        ChannelAccessControlsLong long2 = (ChannelAccessControlsLong) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_CTRL_LONG, UTF8);
        assertNotEquals(long1, long2);
        long2.setValue(intArray);
        long2.setAlarmStatus(alarmStatus);
        long2.setAlarmSeverity(alarmSeverity);
        long2.setUnits(engineeringUnits);
        long2.setUpperDisplayLimit(upperDisplayLimit);
        long2.setLowerDisplayLimit(lowerDisplayLimit);
        long2.setUpperAlarmLimit(upperAlarmLimit);
        long2.setUpperWarningLimit(upperWarningLimit);
        long2.setLowerWarningLimit(lowerWarningLimit);
        long2.setLowerAlarmLimit(lowerAlarmLimit);
        assertNotEquals(long1, long2);
        long2.setUpperControlLimit(upperControlLimit);
        assertNotEquals(long1, long2);
        long2.setLowerControlLimit(lowerControlLimit);
        assertEquals(long1, long2);
        assertEquals(long1.hashCode(), long2.hashCode());
    }

    @Test
    public void createDouble() {
        double[] doubleArray = new double[] { 5, -7, 2174, -35 };
        ChannelAccessDouble double1 = ChannelAccessValueFactory
                .createDouble(doubleArray);
        assertEquals(ChannelAccessValueType.DBR_DOUBLE, double1.getType());
        assertEquals(double1.getValueSize(), 4);
        assertArrayEquals(doubleArray, double1.getValue().array(), 0.0);
        assertEquals(doubleArray[0], double1.getGenericValueElement(0)
                .doubleValue(), 0.0);
        ChannelAccessDouble double2 = (ChannelAccessDouble) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_DOUBLE, UTF8);
        assertNotEquals(double1, double2);
        double2.setValue(doubleArray);
        assertEquals(double1, double2);
        assertEquals(double1.hashCode(), double2.hashCode());
    }

    @Test
    public void createAlarmDouble() {
        double[] doubleArray = new double[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.HIGH;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.NO_ALARM;
        ChannelAccessAlarmDouble double1 = ChannelAccessValueFactory
                .createAlarmDouble(doubleArray, alarmStatus, alarmSeverity);
        assertEquals(ChannelAccessValueType.DBR_STS_DOUBLE, double1.getType());
        assertEquals(alarmStatus, double1.getAlarmStatus());
        assertEquals(alarmSeverity, double1.getAlarmSeverity());
        ChannelAccessAlarmDouble double2 = (ChannelAccessAlarmDouble) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_STS_DOUBLE, UTF8);
        assertNotEquals(double1, double2);
        double2.setValue(doubleArray);
        assertNotEquals(double1, double2);
        double2.setAlarmStatus(alarmStatus);
        double2.setAlarmSeverity(alarmSeverity);
        assertEquals(double1, double2);
        assertEquals(double1.hashCode(), double2.hashCode());
    }

    @Test
    public void createTimeDouble() {
        double[] doubleArray = new double[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        int seconds = 123;
        int nanoseconds = 456;
        ChannelAccessTimeDouble double1 = ChannelAccessValueFactory
                .createTimeDouble(doubleArray, alarmStatus, alarmSeverity,
                        seconds, nanoseconds);
        assertEquals(ChannelAccessValueType.DBR_TIME_DOUBLE, double1.getType());
        assertEquals(seconds, double1.getTimeSeconds());
        assertEquals(nanoseconds, double1.getTimeNanoseconds());
        ChannelAccessTimeDouble double2 = (ChannelAccessTimeDouble) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_TIME_DOUBLE, UTF8);
        assertNotEquals(double1, double2);
        double2.setValue(doubleArray);
        assertNotEquals(double1, double2);
        double2.setAlarmStatus(alarmStatus);
        assertNotEquals(double1, double2);
        double2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(double1, double2);
        double2.setTimeSeconds(seconds);
        assertNotEquals(double1, double2);
        double2.setTimeNanoseconds(nanoseconds);
        assertEquals(double1, double2);
        assertEquals(double1.hashCode(), double2.hashCode());
    }

    @Test
    public void createGraphicsDouble() {
        double[] doubleArray = new double[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        short precision = 3;
        String engineeringUnits = "abcdefgh";
        byte[] rawEngineeringUnits = new byte[] { 97, 98, 99, 100, 101, 102,
                103, 0 };
        double upperDisplayLimit = 1000.0;
        double lowerDisplayLimit = -1000.0;
        double upperAlarmLimit = 456.0;
        double upperWarningLimit = 123.0;
        double lowerWarningLimit = 50.0;
        double lowerAlarmLimit = 1.0;
        ChannelAccessGraphicsDouble double1 = ChannelAccessValueFactory
                .createGraphicsDouble(doubleArray, alarmStatus, alarmSeverity,
                        precision, engineeringUnits, upperDisplayLimit,
                        lowerDisplayLimit, upperAlarmLimit, upperWarningLimit,
                        lowerWarningLimit, lowerAlarmLimit, UTF8);
        assertEquals(ChannelAccessValueType.DBR_GR_DOUBLE, double1.getType());
        assertEquals(precision, double1.getPrecision());
        assertEquals(engineeringUnits, double1.getUnits());
        assertArrayEquals(rawEngineeringUnits, double1.getRawUnits());
        assertEquals(upperDisplayLimit, double1.getUpperDisplayLimit(), 0.0);
        assertEquals(lowerDisplayLimit, double1.getLowerDisplayLimit(), 0.0);
        assertEquals(upperAlarmLimit, double1.getUpperAlarmLimit(), 0.0);
        assertEquals(upperWarningLimit, double1.getUpperWarningLimit(), 0.0);
        assertEquals(lowerWarningLimit, double1.getLowerWarningLimit(), 0.0);
        assertEquals(lowerAlarmLimit, double1.getLowerAlarmLimit(), 0.0);
        ChannelAccessGraphicsDouble double2 = (ChannelAccessGraphicsDouble) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_GR_DOUBLE, UTF8);
        assertNotEquals(double1, double2);
        double2.setValue(doubleArray);
        double2.setAlarmStatus(alarmStatus);
        double2.setAlarmSeverity(alarmSeverity);
        assertNotEquals(double1, double2);
        double2.setPrecision(precision);
        assertNotEquals(double1, double2);
        double2.setUpperDisplayLimit(upperDisplayLimit);
        assertNotEquals(double1, double2);
        double2.setLowerDisplayLimit(lowerDisplayLimit);
        assertNotEquals(double1, double2);
        double2.setUpperAlarmLimit(upperAlarmLimit);
        assertNotEquals(double1, double2);
        double2.setUpperWarningLimit(upperWarningLimit);
        assertNotEquals(double1, double2);
        double2.setLowerWarningLimit(lowerWarningLimit);
        assertNotEquals(double1, double2);
        double2.setLowerAlarmLimit(lowerAlarmLimit);
        assertNotEquals(double1, double2);
        double2.setRawUnits(rawEngineeringUnits);
        assertNotEquals(double1, double2);
        double2.setUnits(engineeringUnits);
        assertEquals(double1, double2);
        assertEquals(double1.hashCode(), double2.hashCode());
    }

    @Test
    public void createControlsDouble() {
        double[] doubleArray = new double[] { 5, -7, 2174, -35 };
        ChannelAccessAlarmStatus alarmStatus = ChannelAccessAlarmStatus.NO_ALARM;
        ChannelAccessAlarmSeverity alarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        short precision = 3;
        String engineeringUnits = "abcdefgh";
        double upperDisplayLimit = 1000.0;
        double lowerDisplayLimit = -1000.0;
        double upperAlarmLimit = 456.0;
        double upperWarningLimit = 123.0;
        double lowerWarningLimit = 50.0;
        double lowerAlarmLimit = 1.0;
        double upperControlLimit = Double.POSITIVE_INFINITY;
        double lowerControlLimit = Double.NaN;
        ChannelAccessControlsDouble double1 = ChannelAccessValueFactory
                .createControlsDouble(doubleArray, alarmStatus, alarmSeverity,
                        precision, engineeringUnits, upperDisplayLimit,
                        lowerDisplayLimit, upperAlarmLimit, upperWarningLimit,
                        lowerWarningLimit, lowerAlarmLimit, upperControlLimit,
                        lowerControlLimit, UTF8);
        assertEquals(ChannelAccessValueType.DBR_CTRL_DOUBLE, double1.getType());
        assertEquals(upperControlLimit, double1.getUpperControlLimit(), 0.0);
        assertEquals(lowerControlLimit, double1.getLowerControlLimit(), 0.0);
        ChannelAccessControlsDouble double2 = (ChannelAccessControlsDouble) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_CTRL_DOUBLE, UTF8);
        assertNotEquals(double1, double2);
        double2.setValue(doubleArray);
        double2.setAlarmStatus(alarmStatus);
        double2.setAlarmSeverity(alarmSeverity);
        double2.setPrecision(precision);
        double2.setUnits(engineeringUnits);
        double2.setUpperDisplayLimit(upperDisplayLimit);
        double2.setLowerDisplayLimit(lowerDisplayLimit);
        double2.setUpperAlarmLimit(upperAlarmLimit);
        double2.setUpperWarningLimit(upperWarningLimit);
        double2.setLowerWarningLimit(lowerWarningLimit);
        double2.setLowerAlarmLimit(lowerAlarmLimit);
        assertNotEquals(double1, double2);
        double2.setUpperControlLimit(upperControlLimit);
        assertNotEquals(double1, double2);
        double2.setLowerControlLimit(lowerControlLimit);
        assertEquals(double1, double2);
        assertEquals(double1.hashCode(), double2.hashCode());
    }

    @Test
    public void createConfigureAcknowledgeTransientAlarms() {
        boolean acknowledgeTransientAlarms = true;
        ChannelAccessConfigureAcknowledgeTransientAlarms ackTransient1 = ChannelAccessValueFactory
                .createConfigureAcknowledgeTransientAlarms(acknowledgeTransientAlarms);
        assertEquals(ChannelAccessValueType.DBR_PUT_ACKT,
                ackTransient1.getType());
        assertEquals(acknowledgeTransientAlarms,
                ackTransient1.isExplicitlyAcknowledgeTransientAlarms());
        assertEquals((short) 1, ackTransient1.getValue().get(0));
        ChannelAccessConfigureAcknowledgeTransientAlarms ackTransient2 = (ChannelAccessConfigureAcknowledgeTransientAlarms) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_PUT_ACKT, UTF8);
        // Number of elements should be one.
        assertEquals(1, ackTransient2.getValueSize());
        assertNotEquals(acknowledgeTransientAlarms,
                ackTransient2.isExplicitlyAcknowledgeTransientAlarms());
        assertNotEquals(ackTransient1, ackTransient2);
        ackTransient2
                .setExplicitlyAcknowledgeTransientAlarms(acknowledgeTransientAlarms);
        assertEquals(ackTransient1, ackTransient2);
        assertEquals(ackTransient1.hashCode(), ackTransient2.hashCode());
    }

    @Test
    public void createAcknowledgeAlarm() {
        ChannelAccessAlarmSeverity acknowledgeSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        ChannelAccessAcknowledgeAlarm ackAlarm1 = ChannelAccessValueFactory
                .createAcknowledgeAlarm(acknowledgeSeverity);
        assertEquals(ChannelAccessValueType.DBR_PUT_ACKS, ackAlarm1.getType());
        assertEquals(acknowledgeSeverity,
                ackAlarm1.getAcknowledgeAlarmSeverity());
        ChannelAccessAcknowledgeAlarm ackAlarm2 = (ChannelAccessAcknowledgeAlarm) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_PUT_ACKS, UTF8);
        // Number of elements should be one.
        assertEquals(1, ackAlarm2.getValueSize());
        assertNotEquals(acknowledgeSeverity,
                ackAlarm2.getAcknowledgeAlarmSeverity());
        assertNotEquals(ackAlarm1, ackAlarm2);
        ackAlarm2.setAcknowledgeAlarmSeverity(acknowledgeSeverity);
        assertEquals(ackAlarm1, ackAlarm2);
        assertEquals(ackAlarm1.hashCode(), ackAlarm2.hashCode());
    }

    @Test
    public void createAlarmAcknowledgementStatus() {
        List<String> stringList = Arrays.asList(new String[] { "abc", "def",
                "123456789012345678901234567890123456789" });
        byte[] rawArray = new byte[] { 97, 98, 99, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 100, 101, 102, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 49, 50, 51, 52, 53, 54, 55, 56, 57, 48,
                49, 50, 51, 52, 53, 54, 55, 56, 57, 48, 49, 50, 51, 52, 53, 54,
                55, 56, 57, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 0 };
        ChannelAccessAlarmStatus currentAlarmStatus = ChannelAccessAlarmStatus.CALC;
        ChannelAccessAlarmSeverity currentAlarmSeverity = ChannelAccessAlarmSeverity.MINOR_ALARM;
        boolean explicitlyAcknowledgeTransientAlarms = true;
        ChannelAccessAlarmSeverity unacknowledgedAlarmSeverity = ChannelAccessAlarmSeverity.MAJOR_ALARM;
        ChannelAccessAlarmAcknowledgementStatus alarmStatus1 = ChannelAccessValueFactory
                .createAlarmAcknowledgementStatus(stringList,
                        currentAlarmStatus, currentAlarmSeverity,
                        explicitlyAcknowledgeTransientAlarms,
                        unacknowledgedAlarmSeverity, UTF8);
        assertEquals(ChannelAccessValueType.DBR_STSACK_STRING,
                alarmStatus1.getType());
        assertEquals(stringList, alarmStatus1.getValue());
        assertArrayEquals(rawArray, alarmStatus1.getRawValue());
        assertEquals(currentAlarmStatus, alarmStatus1.getAlarmStatus());
        assertEquals(currentAlarmSeverity, alarmStatus1.getAlarmSeverity());
        assertEquals(explicitlyAcknowledgeTransientAlarms,
                alarmStatus1.isExplicitlyAcknowledgeTransientAlarms());
        assertEquals(unacknowledgedAlarmSeverity,
                alarmStatus1.getUnacknowledgedAlarmSeverity());
        ChannelAccessAlarmAcknowledgementStatus alarmStatus2 = (ChannelAccessAlarmAcknowledgementStatus) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_STSACK_STRING, UTF8);
        assertNotEquals(alarmStatus1, alarmStatus2);
        alarmStatus2.setAlarmStatus(currentAlarmStatus);
        assertNotEquals(alarmStatus1, alarmStatus2);
        alarmStatus2.setAlarmSeverity(currentAlarmSeverity);
        assertNotEquals(alarmStatus1, alarmStatus2);
        alarmStatus2
                .setExplicitlyAcknowledgeTransientAlarms(explicitlyAcknowledgeTransientAlarms);
        assertNotEquals(alarmStatus1, alarmStatus2);
        alarmStatus2
                .setUnacknowledgedAlarmSeverity(unacknowledgedAlarmSeverity);
        assertNotEquals(alarmStatus1, alarmStatus2);
        alarmStatus2.setRawValue(rawArray);
        assertEquals(alarmStatus1, alarmStatus2);
        assertEquals(alarmStatus1.hashCode(), alarmStatus2.hashCode());
    }

    @Test
    public void createClassName() {
        List<String> stringList = Arrays.asList(new String[] { "abc", "def",
                "123456789012345678901234567890123456789012" });
        byte[] rawArray = new byte[] { 97, 98, 99, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 100, 101, 102, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 49, 50, 51, 52, 53, 54, 55, 56, 57, 48,
                49, 50, 51, 52, 53, 54, 55, 56, 57, 48, 49, 50, 51, 52, 53, 54,
                55, 56, 57, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 0 };
        ChannelAccessClassName className1 = ChannelAccessValueFactory
                .createClassName(stringList, UTF8);
        assertEquals(ChannelAccessValueType.DBR_CLASS_NAME,
                className1.getType());
        assertEquals(3, className1.getValueSize());
        assertEquals(stringList, className1.getValue());
        assertArrayEquals(rawArray, className1.getRawValue());
        assertEquals(className1.getGenericValueElement(1), stringList.get(1));
        ChannelAccessClassName className2 = (ChannelAccessClassName) ChannelAccessValueFactory
                .createValue(ChannelAccessValueType.DBR_CLASS_NAME, UTF8);
        className2.setRawValue(rawArray);
        assertNotEquals(className1, className2);
        className2.setValue(stringList);
        assertEquals(className1, className2);
        assertEquals(className1.hashCode(), className2.hashCode());
    }

}
